#!/usr/bin/perl

# extract.pl W. Rayner (wrayner@well.ox.ac.uk), A. Mahajan (anubha@well.ox.ac.uk) 2015
# 
#
# v1.1 
#  -Added check of number columns in tab file to number of rows in dct file
#  -Added ability to use n_84_0_0 format ids, as well as F.84.0.0
#
# v1.2
#  -Changed column determination from the phenotype header file to checking of 
#   data file header to ensure missing columns are accounted for in the extraction
#
# v1.2.1
#  -Changed the requirement for .sample file, if no sample file order is random
#  -modified split command on data file to only take columns of interest


$| = 1;

use strict;
use warnings;
use Getopt::Long;
use File::Basename;

# file names
my $samplefile = '';                   # sample file with the correct mappings and order, set using -f
my $phenotypefile = '';                # main tab delimited phenotype file to extract data from, could be QC file, set using -p
my $input;                             # input file with column headers (either name or F.22000)
my $output;                            # output file name
my $mapping = "";                      # .dct mapping file for the header names, set using -m
my $relatedfile = "./relateds.txt";    # file of related samples information

my $writeflag = 1;
my @newheader;
my $tempfile = 10000*rand();           # temporary file for extracted results

#hashes/arrays for data
my %headerData;                        # hash of headers from the user file
my @samples;                           # ordered array of the samples 
my $related = 0;                       # flag to indicate inclusion or exclusion of related samples
my %headers;                           # hash of the headers from the dct file
my %mappinghash;                       # hash of the f. id to header name mapping
my %related_samples;                   # hash of the sample ids considered to be related at 2nd degree or above, as defined in $relatedfile
 
GetOptions
 (
 "f|file=s"      => \$samplefile,
 "p|phenotype=s" => \$phenotypefile,
 "i|input=s"     => \$input,
 "o|output=s"    => \$output,
 "r|related"     => \$related,         # setting flag blanks the extracted phenotypes of samples considered to be related
 "m|mapping=s"   => \$mapping
 );

my $columns = getwidth(); 
my $mid = int($columns/2+0.5);
#print "$mid\n";
#print "\n\n           extract.pl\n A phenotype extraction program for\n         UK BioBank data\n\n            W.Rayner\n            A.Mahajan\n              2015\n          Version 1.2.1\n\n";
printf("%*s", $mid+5, "extract.pl\n");
printf("%*s", $mid+25, "A phenotype extraction program for UK BioBank data\n");
printf("%*s", $mid+4, "W.Rayner\n");
printf("%*s", $mid+5, "A.Mahajan\n");
printf("%*s", $mid+2, "2015\n");
printf("%*s", $mid+6, "Version 1.2.1\n");

if (!$input)
 {
 extractusage();
 die "ERROR: No input file specified\n";
 }
if (!$output)
 {
 extractusage();
 die "ERROR: No output file specified\n";
 }

print "Options set:\n";
print "  Sample file:     $samplefile\n";
print "  Phenotype file:  $phenotypefile\n";
print "  Dictionary file: $mapping\n";
print "  Input file:      $input\n";
print "  Output file:     $output\n\n";

if ($related)
 {
 print "Blank related flag set\n";
 %related_samples = get_relateds($relatedfile);
 }

print "Reading phenotypes to extract from: $input\n";
if (-e $input)
 {
 %headerData = read_input($input); 
 print "Done\n\n";
 }
else
 {
 extractusage();
 die "ERROR: input file not found\n"; 
 }

# read dct file for the ids of the given header names and ids (will check positions in the data file later)
print "Reading dictionary file: $mapping\n";
if (-e $mapping)
 {
 %headers = read_dct($mapping, \%headerData); 
 %mappinghash = read_mapping($mapping);
 print "Done\n\n";
 }
else
 {
 extractusage();
 die "ERROR: No dictionary file specified or dictionary file not found\n"; 
 }

my @total_dct = keys %mappinghash;
my @total_tab = get_total_tab($phenotypefile);

if ($#total_dct != $#total_tab)
 {
 print "WARNING: dct file and data file have differing numbers of columns\nWill match on ids to extract data\n";
 }

my @temp_header = keys %headers;

my $n_phenotypes = $#temp_header + 1;
print "Matching input headers to dictionary file\n";
print "Matched $n_phenotypes headers\n\n";

if ($samplefile)
 {
 print "Reading sample order from: $samplefile\n";
 if (-e $samplefile)
  {
  @samples = read_sample($samplefile);
  print "Done\n\n";
  }
 else
  {
  print "Sample file not found, order of samples will not match genotype file\n";
  #changed here so if no sample file found then to use the sample order as the phenotype file
  #die "ERROR: No sample file specified or sample file not found\n"; 
  @samples = read_sample_datafile($phenotypefile);
  }
 } 
else
 {
 print "No sample file specified, order of samples will be the same as phenotype file\n";
 @samples = read_sample_datafile($phenotypefile);
 }

my $n_samples = $#samples + 1;
print "Extracting $n_phenotypes phenotypes for $n_samples samples\n";

if (-e $phenotypefile)
 {
 extract_data($phenotypefile, $tempfile, \@samples, \%headers, \%related_samples, $related);
 write_data($output, $tempfile, \@samples, \%mappinghash);
 my $args = "rm $tempfile";
 system($args);
 }
else
 {
 extractusage();
 print "ERROR: No phenotype file specified or phenotype file not found\n"; 
 }
 
#--------------------------------------------------------------------#

sub extractusage
 {
 print "Usage:\n";
 print " extract.pl -p phenotypefile -m dctfile -i inputfilename -o outputfilename [-f samplefile -r]\n\n";
 print "Options:\n\n";

 #printusage($options, $files, $text);
 printusage("-p --phenotype","phenotypefile","Full path to the file containing the tab delimited phenotype data, with headers in the format f.12345.6.7");
 printusage("-i --input","inputfile","File should contain one row per entry required in the output, format can be any of f.12345.6.7, n_12345_6_7 or the field description, if using the field description all entries that match will be extracted and numbered to match the field name");
 printusage("-o --output","outputfilename","Output file name");
 printusage("-m --mapping","dctfile","Full path to the .dct dictionary file containing the column headers");
 printusage("-f --file","samplefile","Optional, if supplied should be in snptest .sample format and will determine the samples included and order");
 printusage("-r --related","","Optional flag, specifies to keep on non related individuals, requires the accompanying relateds.txt file");      
  
 #print " -p --phenotype\tphenotypefile\tFull path to the file containing the tab delimited phenotype data, with headers in the format f.12345.6.7\n";
 #print " -i --input\tinputfile\tFile should contain one row per entry required in the output, format can be any of f.12345.6.7, n_12345_6_7 or the field description,";
 #print " if using the field description all entries that match will be extracted and numbered to match the field name\n";
 #print " -o --output\toutputfilename\tOutput file name\n";
 #print " -m --mapping\tdctfile\t\tFull path to the .dct dictionary file containing the column headers\n";
 #print " -f --file\tsamplefile\tOptional, if supplied should be in snptest .sample format and will determine the samples included and order\n";
 #print " -r --related\t\t\tOptional flag, specifies to keep on non related individuals, requires the accompanying relateds.txt file\n\n\n";       
 print "\n";
 }


sub getwidth
 {
 my $output = `stty size`;
 my @rowcols = split(/\s/, $output);
 my $cols = $rowcols[1];
 return $cols;
 }

sub printusage
 {
 my $option = $_[0];
 my $file = $_[1];
 my $text = $_[2];
 
 my $cols = getwidth(); 

 printf("%-18s", $option);
 printf("%-15s", $file);
 my $textwidth = $cols - 36; #33 plus 3 spaces
 my $newtext = substr ($text, 0, $textwidth);
 
 if (length($text) > $textwidth)
  {
  printf("  %*s", $textwidth, $newtext);
  for (my $i = $textwidth; $i < length($text); $i+= $textwidth)
   {
   my $newtext = substr ($text, $i, $textwidth);
   $newtext =~ s/^\s//;
   printf("%33s", " ");
   printf("   %*s", -$textwidth, $newtext);
   }
  }
 else
  {
  printf("  %*s", -$textwidth, $newtext);
  }
 print "\n"; 
 }

sub read_sample_datafile
 {
 my $file = $_[0];
 my @samplelist;
 
 open IN, "$file" or die;
 while (<IN>)
  {
  my @temp = split(/\t/, $_, 2);
  push @samplelist, $temp[0];
  }
 close IN; 
 return @samplelist;
 }

sub get_total_tab
 {
 my $file = $_[0];
 open IN, "$file" or die;
 my $header = <IN>;
 close IN;
 
 chomp $header;
 my @temp = split(/\t/, $header);
 return @temp;
 }

sub get_id_column
 {
 my @titles = @{$_[0]};
 #print "@titles\n";
 my $idcolumn = 0;
 for (my $i = 0; $i <= $#titles; $i++)
  {
  if ($titles[$i] eq 'f.eid')
   {
   $idcolumn = $i;
   }
  }
 return $idcolumn; 
 }

sub get_headers_to_extract
 {
 my $line = $_[0];
 my %headers = %{$_[1]};
 my %toextract;
 
 my @titles = split(/\t/, $line);
 for (my $i = 0; $i <= $#titles; $i++)
  {
  if ($headers{$titles[$i]})
   {
   $toextract{$i} = 1;
   }
  }
 return %toextract;
 }

sub get_max_header_number
 {
 my $line = $_[0];
 my %headers = %{$_[1]};
 my $max = 0;
 
 my @titles = split(/\t/, $line);
 for (my $i = 0; $i <= $#titles; $i++)
  {
  if ($headers{$titles[$i]})
   {
   if ($i > $max)
    {
    $max = $i;
    }
   }
  }
 if (!$max)
  {#if by some error there's no match return max as the total number of columns
  $max = $#titles;
  }
 $max+=2; 
 return $max;
 }

sub extract_data
 {
 my $file = $_[0];
 my $tempfile = $_[1];
 my @samples = @{$_[2]};
 my %headers = %{$_[3]};
 my %related = %{$_[4]};
 my $flag = $_[5];
 my $j = 0;
 my %samplehash;
 my %headerToExtract;
 my $max;
 my $line = 0;
 
 print "Opening data file: $file\n";
 open IN, "$file" or die;
 my $headerline = <IN>;
 chomp $headerline;
 
 %headerToExtract = get_headers_to_extract($headerline, \%headers);
 $max = get_max_header_number($headerline, \%headers);
 print "Max columns to split $max\n";
 my @titles = split(/\t/, $headerline);
 
 open TEMP, ">$tempfile" or die $!;
  
 # get id column and add to the list in case it's been left out 
 my $idcolumn = get_id_column(\@titles);
 $headerToExtract{$idcolumn} = 1;
 
 for (my $i = 0; $i <= $#titles; $i++)
  {
  if ($headerToExtract{$i})
   {
   if ($j != 0)
    {
    print TEMP "\t";
    }
   print TEMP "$titles[$i]";
   $j++;
   }
  }
 print TEMP "\n";
 
 print "Reading data\n";
 #print "@samples\n";
 $samplehash{$_}++ for (@samples);
  
 while (<IN>)
  {
  s/\"//g;
  $j = 0;
  chomp;
  my @temp = split(/\t/, $_, $max);
  if ($line % 10000 == 0)
   {
   print "$line\n";
   }
   
  if ($samplehash{$temp[$idcolumn]})
   {
   for (my $i = 0; $i <= $#temp; $i++)
    {
    if ($headerToExtract{$i})
     {
     if ($j != 0)
      {
      print TEMP "\t";
      }
    
     if ($flag and $related{$temp[$idcolumn]} and ($i != $idcolumn))
      {
      print TEMP "NA";
      }
     else
      {
      print TEMP "$temp[$i]";
      }
    
     $j++;
     }
    }
   print TEMP "\n"; 
   }
  $line++;
  }
 print "\nDone\n";
 close IN; 
 close TEMP;
 }

sub write_data
 {
 my $outfile = $_[0];
 my $infile = $_[1];
 my @samplearray = @{$_[2]};
 my %mappinghash = %{$_[3]};
 my %data;
 
 print "Sorting and writing: $outfile\n\n";
 open IN, "$infile" or die $!;
 
 my $title = <IN>;
 chomp $title;
 while (<IN>)
  {
  chomp;
  my @temp = split/\t/;
  $data{$temp[0]} = $_;
  }
 close IN;
 
 open OUT, ">$outfile" or die $!;
 
 my @titles = split(/\t/, $title);
 for (my $i = 0; $i <= $#titles; $i++)
  {
  if ($i != 0)
   {
   print OUT "\t";
   }
  print OUT "$mappinghash{$titles[$i]}";
  }
 print OUT "\n";
 
 for (my $i = 0; $i <= $#samplearray; $i++)
  {
  if ($data{$samplearray[$i]})
   {
   print OUT "$data{$samplearray[$i]}\n";
   }
  }
 close OUT;
 }

sub read_sample
 {
 my $file = $_[0];
 my @sampledata;
 my $i = 0;
 
 open IN, "$file" or die $!;
 while (<IN>)
  {
  chomp;
  my @temp = split/\s+/;
  $sampledata[$i] = $temp[0];
  #$sampledata[$i][1] = $temp[1];
  #$sampledata[$i][2] = $temp[2];
  $i++;
  }
 close IN;
 print "Read $i Samples\n"; 
 return @sampledata; 
 }

# read input file
sub read_input
 {
 my $file = $_[0];
 my %inputData;
 my $count = 0;
 
 open IN, "$file" or die $!;
 while (<IN>)
  {
  chomp;
  my @temp = split/\t/;
  $inputData{$temp[0]} = 1;
  #print "$temp[0]\n";
  $count++;
  }
 close IN; 
 print "Read $count lines\n";
 return %inputData;
 }

sub usage()
 {
 print "Usage:\n";
 }
 
sub read_dct
 {
 my $file = $_[0];
 my %headerList = %{$_[1]};
 my %headers;
 my $j = 0;
 
 open IN, "$file" or die $!;
 
 while (<IN>)
  {
  chomp;
  s/\"//g;
  my @temp = split(/\t/);
  if ($#temp > 1)
   {
   my @ids = split(/\_/, $temp[1]);
   my $id = 'f';
   for (my $i = 1; $i <= $#ids; $i++)
    {
    $id = $id.'.'.$ids[$i];
    }
   
   if ($headerList{$id} or $headerList{$temp[3]} or $headerList{$temp[1]})
    {
    $headers{$id} = 1; #sets headers to f. type id to match the data file
    # print "$headerList{$id} $headerList{$temp[3]}\n";
    }
   $j++; 
   }
  }
 close IN; 
 
 return %headers;
 }
 
sub read_mapping
 {
 my $file = $_[0];
 my %headermapping;
 my $name;
 #open LOG, ">extract.log";

 open IN, "$file" or die $!;
 while (<IN>)
  {
  chomp;
  s/\"//g;
  my @temp = split(/\t/);
  if ($#temp > 2)
   {
   my @ids = split(/\_/, $temp[1]);
   my $id = 'f';
   for (my $i = 1; $i <= $#ids; $i++)
    {
    $id = $id.'.'.$ids[$i];
    }
   
   if ($id =~ /f\.\d+?\.(\d+?)\.(\d+)$/)
    {
    my $fieldNum = $2;# need to take both fields for id
    my $field = $1;
    if ($fieldNum > 0 or $field > 0)
     {
     $name = $temp[3].'_'.$field.'_'.$fieldNum;
     }
    else
     {
     $name = $temp[3];
     }
    }
   else
    {
    $name = $temp[3];
    }
   $headermapping{$id} = $name;
   #print LOG "$id\t$name\n";

   }
  }
 close IN; 
   # close LOG;
 
 return %headermapping;
 } 

 
sub get_relateds
 {
 my $file = $_[0];
 my %related;
 my %sample;
 my $related_cut_off = 0.0884;
 open IN, "$file" or die; 
 open LOG, ">related-removed.log" or die $!;
 
 while (<IN>)
  {
  # count the number of times each sample is seen in the file (at 2nd degree relatedness or higher)
  chomp;
  my @temp = split/\t/;
  if ($temp[2] > $related_cut_off)
   {
   $sample{$temp[0]}++;
   $sample{$temp[1]}++;
   }
  }
 close IN;
 
 open IN, "$file"; 
 while (<IN>)
  {
  chomp;
  my @temp = split/\t/;
  if ($temp[2] > $related_cut_off)
   {
   if ($sample{$temp[1]} > 1)
    {
    $related{$temp[1]} = 1;
    }
   else
    {
    $related{$temp[0]} = 1;
    }
   }
  } 
 my @toremove = keys %related;
 foreach my $sample (@toremove)
  {
  print LOG "$sample\n";
  }
 close LOG; 
 my $total = $#toremove+1;
 print "Will blank $total Samples as related\nSee related-removed.log for ids\n\n";
 return %related;
 }
 
#---------------------------------------------------------------------# 
